#include <vector>
#include <iostream>
#include "ndominated.h"
#include "VectorSet.h"

using namespace std;


/*
 * This methods finds the non-dominated elements from a given set of vectors using the 
 * algorithm developed by Kung, Luccio and Preparata. This algorithm has time complexity
 * O(n lg n) for k = 2, and O(n (lg^{k-2} n)) for k >= 3, where n is the size of the input
 * vector set and k is the dimension of the vectors.
 * This method was implemented following the directions given by Bentley in:
 *   Jon Louis Bentley, Multidimensional divide-and-conquer, Communications of the ACM, 
 *   23 (1980), no. 4, 214-229.
 * 
 * 
 */ 
void NonDominance::kungNDominatedSet(vector<double> vSet[], int size, VectorSet &nDominated) {
   // Sort vector set with respect to the first component.
	myQuickSort(vSet, 0, size-1);
	
	// Solve the problem recursively dividing the set into two halves.
	minimalSet(vSet, nDominated, 0, size-1);	
}

/*
 * This method divide the vector set into two halves and then solve them recursively.
 */
void NonDominance::minimalSet(vector<double> v[], VectorSet &r, int first, int last) {
	VectorSet s;
	int m;

	/* If the dimension of vectors is 2 */
	if (v[first].size() == 2) {
		minimalSet2D(v, r, last);
		return;
	}

	/* When the vector set has only one element, return this element as the nondominated
	 * element (minimal element) 
	 */
	if ( first >= last ) {
		r.addElement(v[first]);
		return;
	}
	
	/* Divide recursively to solve two problems of size n/2 */
	m = (first + last) / 2;
	
	/* r and s are the nondominated set of each half */
	minimalSet(v, r, first, m);
	minimalSet(v, s, m + 1, last);

	/* The whole set r belongs to the global nondominated set, however we have 
	 * to check which elements of s are global nondominated vectors. 
	 */
	unionM(r, s);	
}


void NonDominance::minimalSet2D(vector<double> v[], VectorSet &r, int last) {
	double minimum;
	
	minimum = v[0].at(1);
	r.addElement(v[0]);	
	
	for (int i = 1; i < last; ++i)
		if (v[i].at(1) < minimum) 
		{
		   minimum = v[i].at(1);
			r.addElement(v[i]);
		}
}

void NonDominance::unionM(VectorSet &rSet, VectorSet sSet) {
	list< vector<double> >::iterator is, jr;
	bool dominated;
		
	for (is = sSet.begin(); is != sSet.end(); ++is) {
		
		dominated = false;
		jr = rSet.begin();
		
		while (jr != rSet.end() && !dominated) {
			if (dominates(1, *jr, *is))
				dominated = true;
			else if (dominates(0, *is, *jr)) {
				jr = rSet.deleteElement(jr);								
			}
			else
				++jr;
		}
		
		if (!dominated)
			rSet.addElement(*is);		
	}	
}

void NonDominance::myQuickSort(vector<double> A[], int p, int r) {
	int q;
	
	if (p < r) {
		q = partition(A, p, r);
		
		myQuickSort(A, p, q-1);
		myQuickSort(A, q+1, r);
	}
}

int NonDominance::partition(vector<double> A[], int p, int r) {
	int i, j;
	vector<double> x;
	vector<double> hold(A[1].size(), 0);
	
	x = A[r];
	i = p - 1;
		
	for (j = p; j <= r-1; ++j) {
		if (A[j].at(0) < x.at(0)) {
			++i;
			
			hold = A[i];
			A[i] = A[j];
			A[j] = hold;
		}
		else if (A[j].at(0) == x.at(0)) {
			if (breakTie(A[j], x) <= 0) {			
				++i;
			
				hold = A[i];
				A[i] = A[j];
				A[j] = hold;
			}
		}
	}
	
	hold = A[i+1];
	A[i+1] = A[r];
	A[r] = hold;
	
	return i+1;
}

inline int NonDominance::breakTie(vector<double> &a, vector<double> &b) {
	for (unsigned i = 1; i < a.size(); ++i) {
		if (a[i] < b[i])
			return -1;
		else if (a[i] > b[i])
			return 1;
	}
	
	return 0;
}


/**
 * Method that checks if vector x dominates vector y.
 */
bool NonDominance::dominates(int first, vector<double> x, vector<double> y) {
	unsigned counter = 0;
	
	for (unsigned int i = first; i < x.size(); ++i) {
		if (x[i] > y[i])
			return false;
		else if (x[i] == y[i])
			++counter;		
	}
	
	if (counter == x.size())
		return false;
	else
		return true;
}

